//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/Model/Data/Data2DItem.cpp
//! @brief     Implements class Data2DItem.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/Model/Data/Data2DItem.h"
#include "Base/Axis/Frame.h"
#include "Base/Axis/Scale.h"
#include "Base/Util/Assert.h"
#include "Device/Data/Datafield.h"
#include "Device/Mask/IShape2D.h"
#include "Device/Mask/MaskStack.h"
#include "GUI/Model/Axis/AmplitudeAxisItem.h"
#include "GUI/Model/Axis/BasicAxisItem.h"
#include "GUI/Model/Mask/MaskItems.h"
#include "GUI/Model/Mask/MaskeditorListmodel.h"
#include "GUI/Model/Mask/ProjectionList.h"
#include "GUI/Support/Style/QCP_Util.h"
#include "GUI/Support/XML/UtilXML.h"
#include <qcustomplot.h>

namespace {

namespace Tag {

const QString BaseData("BaseData");
const QString Gradient("Gradient");
const QString Interpolation("Interpolation");
const QString MaskContainer("MaskContainer");
const QString ProjectionContainer("ProjectionContainer");
const QString ZAxis("ZAxis");

} // namespace Tag

const QCPColorGradient infernoGradient = GUI::QCP_Util::colorGradientInferno();

// gradient map for colormaps
const QMap<QString, QCPColorGradient::GradientPreset> gradient_map = {
    {"Grayscale", QCPColorGradient::gpGrayscale},
    {"Hot", QCPColorGradient::gpHot},
    {"Cold", QCPColorGradient::gpCold},
    {"Night", QCPColorGradient::gpNight},
    {"Candy", QCPColorGradient::gpCandy},
    {"Geography", QCPColorGradient::gpGeography},
    {"Ion", QCPColorGradient::gpIon},
    {"Thermal", QCPColorGradient::gpThermal},
    {"Polar", QCPColorGradient::gpPolar},
    {"Spectrum", QCPColorGradient::gpSpectrum},
    {"Jet", QCPColorGradient::gpJet},
    {"Hues", QCPColorGradient::gpHues}};

const QMap<QString, QCPColorGradient> custom_gradient_map = {{"Inferno", infernoGradient}};

const QString startGradient = "Inferno";

} // namespace

Data2DItem::Data2DItem()
    : DataItem(M_TYPE)
    , m_isInterpolated(true)
    , m_gradient(
          ComboProperty::fromList(gradient_map.keys() + custom_gradient_map.keys(), startGradient))
    , m_zAxis(std::make_unique<AmplitudeAxisItem>())
{
}

Data2DItem::~Data2DItem() = default;

void Data2DItem::setDatafield(const Datafield& data)
{
    ASSERT(data.rank() == 2);
    DataItem::setDatafield(data);
    updateAxesZoomLevel();
    updateDataRange();
}

double Data2DItem::xMin() const
{
    const double defaultXmin(0.0);
    return m_datafield ? m_datafield->axis(0).min() : defaultXmin;
}

double Data2DItem::xMax() const
{
    const double defaultXmax(1.0);
    return m_datafield ? m_datafield->axis(0).max() : defaultXmax;
}

double Data2DItem::yMin() const
{
    const double defaultYmin(0.0);
    return m_datafield ? m_datafield->axis(1).min() : defaultYmin;
}

double Data2DItem::yMax() const
{
    const double defaultYmax(1.0);
    return m_datafield ? m_datafield->axis(1).max() : defaultYmax;
}

double Data2DItem::lowerZ() const
{
    return zAxisItem()->min();
}

double Data2DItem::upperZ() const
{
    return zAxisItem()->max();
}

void Data2DItem::setZrange(double zmin, double zmax)
{
    if (lowerZ() == zmin && upperZ() == zmax)
        return;
    zAxisItem()->setMin(zmin);
    zAxisItem()->setMax(zmax);
    emit itemAxesRangeChanged();
}

void Data2DItem::copyZRangeFromItem(DataItem* sourceItem)
{
    const auto* source = dynamic_cast<Data2DItem*>(sourceItem);
    if (!source || source == this)
        return;
    setZrange(source->lowerZ(), source->upperZ());
}

bool Data2DItem::isZoomed() const
{
    return lowerX() > xMin() || upperX() < xMax() || lowerY() > yMin() || upperY() < yMax();
}

QCPColorGradient Data2DItem::currentGradientQCP() const
{
    if (gradient_map.contains(currentGradient()))
        return gradient_map.value(currentGradient());
    return custom_gradient_map.value(currentGradient());
}

QString Data2DItem::currentGradient() const
{
    return gradientCombo().currentValue();
}

void Data2DItem::setCurrentGradient(const QString& gradient)
{
    m_gradient.setCurrentValue(gradient);
    emit gradientChanged();
}
bool Data2DItem::isLog() const
{
    return zAxisItem()->isLogScale();
}

void Data2DItem::setLog(bool islog)
{
    zAxisItem()->setLogScale(islog);
}
void Data2DItem::setInterpolated(bool interp)
{
    m_isInterpolated = interp;
    emit interpolationChanged(interp);
}

bool Data2DItem::isZaxisLocked() const
{
    return m_zAxis->isLocked();
}

void Data2DItem::setZaxisLocked(bool state)
{
    return m_zAxis->setLocked(state);
}

std::vector<int> Data2DItem::axdims() const
{
    return {xSize(), ySize()};
}

//! Sets zoom range of X,Y axes, if it was not yet defined.

void Data2DItem::updateAxesZoomLevel()
{
    // set zoom range of x-axis to min, max values if it was not set already
    if (upperX() < lowerX())
        setXrange(xMin(), xMax());

    // set zoom range of y-axis to min, max values if it was not set already
    if (upperY() < lowerY())
        setYrange(yMin(), yMax());

    const int nx = static_cast<int>(m_datafield->axis(0).size());
    axItemX()->resize(nx);
    const int ny = static_cast<int>(m_datafield->axis(1).size());
    axItemY()->resize(ny);
}

//! Sets min,max values for z-axis, if axes is not locked, and ranges are not yet set.

void Data2DItem::updateDataRange()
{
    if (isZaxisLocked())
        return;

    computeDataRange();
    emit alignRanges();
}

void Data2DItem::computeDataRange()
{
    setZrange(dataRange().first, dataRange().second);
}

//! Init zmin, zmax to match the intensity values range.
QPair<double, double> Data2DItem::dataRange() const
{
    const Datafield* data = c_field();

    const auto vec = data->flatVector();
    double min(*std::min_element(vec.cbegin(), vec.cend()));
    double max(*std::max_element(vec.cbegin(), vec.cend()));
    double logRange = pow(10, zAxisItem()->logRangeOrders());
    if (isLog()) {
        min = std::max(min, max / logRange);
        max *= 1.1;
    } else
        max *= 1.1;

    return {min, max};
}

const AmplitudeAxisItem* Data2DItem::zAxisItem() const
{
    return m_zAxis.get();
}

AmplitudeAxisItem* Data2DItem::zAxisItem()
{
    return m_zAxis.get();
}

void Data2DItem::resetView()
{
    if (!m_datafield)
        return;

    setAxesRangeToData();
    if (!isZaxisLocked())
        computeDataRange();
}

void Data2DItem::setAxesRangeToData()
{
    setXrange(xMin(), xMax());
    setYrange(yMin(), yMax());
}

void Data2DItem::writeTo(QXmlStreamWriter* w) const
{
    XML::writeAttribute(w, XML::Attrib::version, uint(1));

    // parameters from base class
    w->writeStartElement(Tag::BaseData);
    DataItem::writeTo(w);
    w->writeEndElement();

    // interpolation
    w->writeStartElement(Tag::Interpolation);
    XML::writeAttribute(w, XML::Attrib::value, m_isInterpolated);
    w->writeEndElement();

    // gradient
    w->writeStartElement(Tag::Gradient);
    m_gradient.writeTo(w);
    w->writeEndElement();

    // z axis
    w->writeStartElement(Tag::ZAxis);
    m_zAxis->writeTo(w);
    w->writeEndElement();

    // masks and projections
    if (m_mask_model) {
        if (auto* mi = m_mask_model->maskContItem()) {
            w->writeStartElement(Tag::MaskContainer);
            mi->writeTo(w);
            w->writeEndElement();
        }
    }

    // projections
    if (m_proj_model) {
        if (auto* pi = m_proj_model->projnItem()) {
            w->writeStartElement(Tag::ProjectionContainer);
            pi->writeTo(w);
            w->writeEndElement();
        }
    }
}

void Data2DItem::readFrom(QXmlStreamReader* r)
{
    const uint version = XML::readUIntAttribute(r, XML::Attrib::version);
    Q_UNUSED(version)

    while (r->readNextStartElement()) {
        QString tag = r->name().toString();

        // parameters from base class
        if (tag == Tag::BaseData) {
            DataItem::readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // interpolation
        } else if (tag == Tag::Interpolation) {
            XML::readAttribute(r, XML::Attrib::value, &m_isInterpolated);
            XML::gotoEndElementOfTag(r, tag);

            // gradient
        } else if (tag == Tag::Gradient) {
            m_gradient.readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // z axis
        } else if (tag == Tag::ZAxis) {
            m_zAxis->readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // masks
        } else if (tag == Tag::MaskContainer) {
            getOrCreateMaskModel()->maskContItem()->readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

            // projections
        } else if (tag == Tag::ProjectionContainer) {
            getOrCreateProjectionModel()->projnItem()->readFrom(r);
            XML::gotoEndElementOfTag(r, tag);

        } else
            r->skipCurrentElement();
    }
}

MaskList* Data2DItem::maskContainerItem()
{
    return m_mask_model ? m_mask_model->maskContItem() : nullptr;
}

const MaskList* Data2DItem::maskContainerItem() const
{
    return m_mask_model ? m_mask_model->maskContItem() : nullptr;
}

MaskeditorListmodel* Data2DItem::getOrCreateMaskModel()
{
    if (!m_mask_model)
        m_mask_model = std::make_unique<MaskeditorListmodel>(new MaskList);

    return m_mask_model.get();
}

Datafield* Data2DItem::createMaskedField() const
{
    // Requesting mask information
    std::unique_ptr<IShape2D> roi;
    Datafield* result = c_field()->clone();
    MaskStack detectorMask;
    const QVector<const MaskItem*> maskItems = maskContainerItem()->maskItems();
    for (auto maskIter = maskItems.rbegin(); maskIter != maskItems.rend(); maskIter++)
        if ((*maskIter)->isVisible()) {
            if (auto* roiItem = dynamic_cast<const RegionOfInterestItem*>((*maskIter)))
                roi = roiItem->createShape();
            else {
                std::unique_ptr<IShape2D> shape((*maskIter)->createShape());
                detectorMask.pushMask(*shape, (*maskIter)->maskValue(), false);
            }
        }

    // ROI mask has to be the last one, it can not be "unmasked" by other shapes
    if (roi)
        detectorMask.pushMask(*roi, true, false);

    if (!detectorMask.hasMasks())
        return nullptr;

    for (size_t i = 0; i < result->size(); ++i)
        if (detectorMask.isMasked(i, result->frame()))
            (*result)[i] = 0;

    return result;
}

ProjectionList* Data2DItem::projectionContainerItem()
{
    return m_proj_model ? m_proj_model->projnItem() : nullptr;
}

const ProjectionList* Data2DItem::projectionContainerItem() const
{
    return m_proj_model ? m_proj_model->projnItem() : nullptr;
}

MaskeditorListmodel* Data2DItem::getOrCreateProjectionModel()
{
    if (!m_proj_model)
        m_proj_model = std::make_unique<MaskeditorListmodel>(new ProjectionList);

    return m_proj_model.get();
}

bool Data2DItem::hasMasks() const
{
    return (maskContainerItem() && !maskContainerItem()->isEmpty());
}

bool Data2DItem::hasProjections() const
{
    return (projectionContainerItem() && !projectionContainerItem()->isEmpty());
}

void Data2DItem::clearMasks()
{
    if (auto* mask_cont = maskContainerItem())
        mask_cont->clear();
    if (auto* proj_cont = projectionContainerItem())
        proj_cont->clear();
}
