//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Frame/MaskEditingFrame.cpp
//! @brief     Implements class MaskEditingFrame.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Frame/MaskEditingFrame.h"
#include "Base/Util/Assert.h"
#include "GUI/Model/Data/Data2DItem.h"
#include "GUI/Model/Mask/MaskeditorListmodel.h"
#include "GUI/View/Canvas/MaskEditorActions.h"
#include "GUI/View/Canvas/MaskEditorCanvas.h"
#include "GUI/View/Scene/MaskGraphicsScene.h"
#include "GUI/View/Setup/MaskPanel.h"
#include "GUI/View/Setup/MaskToolbar.h"
#include <QSplitter>
#include <QVBoxLayout>

MaskEditingFrame::MaskEditingFrame(QWidget* parent)
    : DataAccessWidget(parent)
    , m_actions(new MaskEditorActions(this))
    , m_panel(new MaskPanel)
    , m_canvas(new MaskEditorCanvas)
{
    auto* layout = new QHBoxLayout;
    setLayout(layout);
    layout->setContentsMargins(0, 0, 0, 0);
    layout->setSpacing(0);

    auto* splitter = new QSplitter;
    layout->addWidget(splitter);
    splitter->setOrientation(Qt::Horizontal);
    splitter->addWidget(m_canvas);
    splitter->addWidget(m_panel);
    splitter->setCollapsible(1, true);

    auto* toolbar = new MaskToolbar(m_actions);
    layout->addWidget(toolbar);

    m_panel->setPanelHidden(true);

    //... Connections

    connect(m_actions, &MaskEditorActions::resetViewRequest, m_canvas,
            &MaskEditorCanvas::onResetViewRequest);

    connect(m_actions, &MaskEditorActions::propertyPanelRequest,
            [this] { m_panel->setPanelHidden(!m_panel->isHidden()); });

    connect(m_actions, &MaskEditorActions::savePlotRequest, m_canvas,
            &MaskEditorCanvas::onSavePlotRequest);

    connect(toolbar, &MaskToolbar::activityChanged, m_canvas->getScene(),
            &MaskGraphicsScene::onActivityChanged);

    connect(toolbar, &MaskToolbar::maskValueChanged, m_canvas->getScene(),
            &MaskGraphicsScene::onMaskValueChanged);

    connect(toolbar, &MaskToolbar::presentationChange, m_canvas,
            &MaskEditorCanvas::onPresentationChange);

    // space bar push (request for zoom mode)
    connect(m_canvas, &MaskEditorCanvas::changeActivityRequest, toolbar,
            &MaskToolbar::onChangeActivityRequest);

    connect(m_canvas, &MaskEditorCanvas::deleteSelectedRequest, m_actions,
            &MaskEditorActions::onDeleteMaskAction);

    connect(m_canvas->getScene(), &MaskGraphicsScene::itemContextMenuRequest, m_actions,
            &MaskEditorActions::onItemContextMenuRequest);

    connect(m_panel, &MaskPanel::itemContextMenuRequest, m_actions,
            &MaskEditorActions::onItemContextMenuRequest);
}

void MaskEditingFrame::setJobOrDatafileItem(QObject* item)
{
    setBaseItem(item);
    updateFrame();
}

//! Returns list of actions intended for styled toolbox (on the top).

QList<QAction*> MaskEditingFrame::actionList()
{
    return m_actions->topToolboxActions();
}

void MaskEditingFrame::updateFrame()
{
    ASSERT(currentData2DItem());

    MaskeditorListmodel* mask_list_model = currentData2DItem()->getOrCreateMaskModel();
    ASSERT(mask_list_model);

    m_panel->updateMaskPanel(mask_list_model);

    m_canvas->updateMaskCanvas(currentData2DItem(), m_panel->selectionModel());

    m_actions->setModels(mask_list_model, m_panel->selectionModel());

    update();
}

void MaskEditingFrame::showEvent(QShowEvent*)
{
    if (currentData2DItem())
        updateFrame();
}

void MaskEditingFrame::hideEvent(QHideEvent*)
{
    m_panel->resetMaskPanel();
    m_canvas->resetMaskCanvas();
}
