//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Frame/Plot2DWithProjectionFrame.cpp
//! @brief     Defines class Plot2DWithProjectionFrame.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Frame/Plot2DWithProjectionFrame.h"
#include "Base/Util/Assert.h"
#include "GUI/Model/Data/Data2DItem.h"
#include "GUI/View/Canvas/ProjectionsEditorCanvas.h"
#include "GUI/View/Canvas/ProjectionsWidget.h"
#include "GUI/View/Setup/ProjectionActions.h"
#include "GUI/View/Setup/ProjectionToolbar.h"
#include "GUI/View/Setup/Scale2DEditor.h"
#include <QHBoxLayout>
#include <QSplitter>

Plot2DWithProjectionFrame::Plot2DWithProjectionFrame(QWidget* parent)
    : DataAccessWidget(parent)
    , m_actions(new ProjectionActions(this))
    , m_projectionsCanvas(new ProjectionsEditorCanvas)
    , m_projectionsWidget(new ProjectionsWidget)
    , m_scaleEditor(new Scale2DEditor(this))
{
    auto* layout = new QHBoxLayout;
    setLayout(layout);
    layout->setContentsMargins(0, 0, 0, 0);
    layout->setSpacing(0);

    auto* splitter = new QSplitter;
    layout->addWidget(splitter);
    splitter->setOrientation(Qt::Vertical);
    splitter->addWidget(m_projectionsCanvas);
    splitter->addWidget(m_projectionsWidget);

    layout->addWidget(m_scaleEditor);
    m_scaleEditor->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Minimum);
    m_scaleEditor->setHidden(true);

    auto* toolbar = new ProjectionToolbar(m_actions);
    layout->addWidget(toolbar);

    //... Connections

    connect(m_actions, &ProjectionActions::resetViewRequest, m_projectionsCanvas,
            &ProjectionsEditorCanvas::onResetViewRequest);

    connect(m_actions, &ProjectionActions::propertyPanelRequest,
            [this] { m_scaleEditor->setHidden(!m_scaleEditor->isHidden()); });

    connect(toolbar, &ProjectionToolbar::activityChanged, m_projectionsCanvas,
            &ProjectionsEditorCanvas::setProjectionsCanvasMode);

    connect(toolbar, &ProjectionToolbar::activityChanged, m_projectionsWidget,
            &ProjectionsWidget::onActivityChanged);

    connect(m_projectionsWidget, &ProjectionsWidget::changeActivityRequest, toolbar,
            &ProjectionToolbar::onProjectionTabChange);

    connect(m_projectionsCanvas, &ProjectionsEditorCanvas::deleteSelectedRequest, m_actions,
            &ProjectionActions::onDeleteAction);

    connect(m_projectionsCanvas, &ProjectionsEditorCanvas::changeProjectionsTabRequest,
            m_projectionsWidget, &ProjectionsWidget::onActivityChanged);

    // space bar push (request for zoom mode)
    connect(m_projectionsCanvas, &ProjectionsEditorCanvas::changeActivityRequest, toolbar,
            &ProjectionToolbar::onChangeActivityRequest);

    connect(m_projectionsCanvas, &ProjectionsEditorCanvas::marginsChanged, m_projectionsWidget,
            &ProjectionsWidget::onMarginsChanged);

    toolbar->onChangeActivityRequest(MaskFlags::HORIZONTAL_LINE_MODE);
}

void Plot2DWithProjectionFrame::setJobOrDatafileItem(QObject* item)
{
    setBaseItem(item);
    m_scaleEditor->setJobOrDatafileItem(item);
    updateFrame();
}

QList<QAction*> Plot2DWithProjectionFrame::actionList()
{
    return m_actions->topToolboxActions();
}

void Plot2DWithProjectionFrame::updateFrame()
{
    ASSERT(currentData2DItem());

    // The following line must come before 'm_projectionsCanvas->setContext()'!
    // It prevents problem with switching between datasets in projection mode.
    m_projectionsWidget->disconnectItem();

    m_projectionsCanvas->updateProjsCanvas(currentData2DItem());
    m_projectionsWidget->setData2DItem(currentData2DItem());

    m_actions->updateProjsActions(currentData2DItem());
}

void Plot2DWithProjectionFrame::showEvent(QShowEvent*)
{
    if (currentData2DItem())
        updateFrame();
}

void Plot2DWithProjectionFrame::hideEvent(QHideEvent*)
{
    m_projectionsCanvas->resetProjsCanvas();
    m_projectionsWidget->disconnectItem();
    m_projectionsWidget->setData2DItem(nullptr);
}
