//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Instrument/InstrumentLibraryDialog.cpp
//! @brief     Implements class InstrumentLibraryDialog.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Instrument/InstrumentLibraryDialog.h"
#include "Base/Util/Assert.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/Model/Device/InstrumentLibrary.h"
#include "GUI/Support/Style/Style.h"
#include "GUI/View/Item/ItemViewOverlayButtons.h"
#include "GUI/View/Tool/ItemDelegateForHTML.h"
#include "GUI/View/Tool/mainwindow_constants.h"
#include "GUI/View/Widget/ApplicationSettings.h"
#include <QAction>
#include <QFormLayout>
#include <QInputDialog>
#include <QPushButton>
#include <QSettings>
#include <QTextEdit>
#include <QVBoxLayout>

InstrumentLibraryDialog::InstrumentLibraryDialog(QWidget* parent,
                                                 InstrumentLibrary* instrumentLibrary)
    : QDialog(parent)
    , m_instrumentLibrary(instrumentLibrary)
    , m_treeModel(new InstrumentsTreeModel(this, instrumentLibrary->instrumentModel()))
    , m_chosenItem(nullptr)
{
    // object name is needed to reload application settings
    if (this->objectName().isEmpty())
        this->setObjectName("InstrumentLibraryDialog");

    setGeometry(0, 0, 320, 429);

    auto* layout = new QVBoxLayout(this);
    setLayout(layout);

    m_treeView = new QTreeView;
    layout->addWidget(m_treeView);

    m_buttonBox = new QDialogButtonBox;
    layout->addWidget(m_buttonBox);
    m_buttonBox->setOrientation(Qt::Horizontal);
    m_buttonBox->setStandardButtons(QDialogButtonBox::Cancel | QDialogButtonBox::Ok);

    setWindowIcon(QIcon(":/images/library.svg"));
    setWindowFlag(Qt::WindowContextHelpButtonHint, false);

    m_treeModel->enableEmptyHeadlines(false);

    m_treeView->setItemsExpandable(false);
    m_treeView->setRootIsDecorated(false);
    m_treeView->setHeaderHidden(true);
    m_treeView->setContextMenuPolicy(Qt::CustomContextMenu);
    m_treeView->setModel(m_treeModel);
    m_treeView->expandAll();
    m_treeView->setVerticalScrollMode(QTreeView::ScrollPerPixel);
    m_treeView->setIndentation(0);
    m_treeView->setItemDelegate(new ItemDelegateForHTML(this));
    m_treeView->setIconSize(QSize(128, 128));

    connect(m_treeModel, &QAbstractItemModel::modelReset, [this] { m_treeView->expandAll(); });

    connect(m_buttonBox, &QDialogButtonBox::accepted, this, &QDialog::accept);
    connect(m_buttonBox, &QDialogButtonBox::rejected, this, &QDialog::reject);

    // ensure a current item when widget is shown
    GUI::Style::setResizable(this);
    appSettings->loadWindowSizeAndPos(this);
}

InstrumentLibraryDialog::~InstrumentLibraryDialog()
{
    appSettings->saveWindowSizeAndPos(this);
}

InstrumentItem* InstrumentLibraryDialog::execChoose()
{
    setWindowTitle("Instrument Library - Choose instrument");

    ItemViewOverlayButtons::install(
        m_treeView, [this](const QModelIndex& i, bool h) { return getOverlayActions(i, h); });
    m_treeView->setItemDelegate(new ItemDelegateForHTML(this));

    connect(m_treeView, &QTreeView::doubleClicked, this,
            &InstrumentLibraryDialog::onItemDoubleClickedForChoose);
    connect(m_treeView->selectionModel(), &QItemSelectionModel::currentChanged, this,
            &InstrumentLibraryDialog::onCurrentChangedForChoose);
    onCurrentChangedForChoose();

    // select the first instrument
    QModelIndex index = m_treeModel->indexForItem(m_instrumentLibrary->instrumentItems().first());
    m_treeView->setCurrentIndex(index);

    if (exec() == QDialog::Accepted)
        return m_chosenItem;

    return nullptr;
}

void InstrumentLibraryDialog::execAdd(const InstrumentItem& instrumentToAdd)
{
    const QString& newName = m_instrumentLibrary->suggestName(instrumentToAdd.instrumentName());
    auto* addedInstrument = m_instrumentLibrary->addItemCopy(newName, instrumentToAdd);

    setWindowTitle("Instrument Library - Add instrument");

    m_treeModel->setNewInstrument(addedInstrument);

    ItemViewOverlayButtons::install(
        m_treeView, [this](const QModelIndex& i, bool h) { return getOverlayActions(i, h); });
    m_treeView->setItemDelegate(new ItemDelegateForHTML(this));

    QModelIndex index = m_treeModel->indexForItem(addedInstrument);
    m_treeView->expandAll();
    m_treeView->setCurrentIndex(index);
    m_treeView->scrollTo(index, QAbstractItemView::PositionAtTop);
    exec();
}

void InstrumentLibraryDialog::onItemDoubleClickedForChoose(const QModelIndex& index)
{
    m_chosenItem = m_treeModel->itemForIndex(index);
    if (m_chosenItem != nullptr)
        accept();
}

void InstrumentLibraryDialog::onCurrentChangedForChoose()
{
    m_chosenItem = m_treeModel->itemForIndex(m_treeView->currentIndex());
    m_buttonBox->button(QDialogButtonBox::Ok)->setEnabled(m_chosenItem != nullptr);
}

QList<QAction*> InstrumentLibraryDialog::getOverlayActions(const QModelIndex& index, bool asHover)
{
    if (m_treeModel->isHeadline(index))
        return {};

    // -- index belongs to item
    if (!asHover)
        return {};

    auto* item = m_treeModel->itemForIndex(index);
    if (item == nullptr)
        return {};

    auto* removeAction = new QAction(this);
    removeAction->setText("Remove");
    removeAction->setIcon(QIcon(":/images/delete.svg"));
    removeAction->setIconText("Remove");
    removeAction->setToolTip("Remove this instrument");
    connect(removeAction, &QAction::triggered, [this, item] { m_treeModel->removeItem(item); });

    return {removeAction};
}

void InstrumentLibraryDialog::onInstrumentNameEdited(const QString& newName)
{
    QModelIndex index = m_treeView->currentIndex();
    m_treeModel->setData(index, newName, Qt::EditRole);
}

void InstrumentLibraryDialog::onInstrumentDescriptionEdited(const QString& t)
{
    QModelIndex index = m_treeView->currentIndex();
    m_treeModel->setData(index, t, Qt::ToolTipRole);
}
