//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Setup/MaskToolbar.cpp
//! @brief     Implements class MaskToolbar.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Setup/MaskToolbar.h"
#include "Base/Util/Assert.h"
#include "GUI/View/Canvas/MaskEditorActions.h"
#include "GUI/View/Tool/mainwindow_constants.h"
#include <QLabel>
#include <QRadioButton>
#include <QStyle>
#include <QToolButton>
#include <QVariant>

MaskToolbar::MaskToolbar(MaskEditorActions* actions)
    : m_activityButtonGroup(new QButtonGroup(this))
{
    setOrientation(Qt::Vertical);
    setStyleSheet("border-left:1px solid rgb(180,180,180)");
    setIconSize(QSize(GUI::Style::toolbar_icon_size, GUI::Style::toolbar_icon_size));

    //... Zoom / selection mode

    auto* panButton = new QToolButton(this);
    panButton->setIcon(QIcon(":/images/hand-right.svg"));
    panButton->setToolTip("Pan/zoom mode (space)");
    panButton->setCheckable(true);
    panButton->setChecked(true);
    addWidget(panButton);

    auto* resetViewButton = new QToolButton(this);
    resetViewButton->setIcon(QIcon(":/images/camera-metering-center.svg"));
    resetViewButton->setToolTip("Reset pan/zoom to initial state");
    addWidget(resetViewButton);
    connect(resetViewButton, &QToolButton::clicked, actions, &MaskEditorActions::resetViewRequest);

    add_separator();

    auto* selectionButton = new QToolButton(this);
    selectionButton->setIcon(QIcon(":/images/maskeditor_arrow.svg"));
    selectionButton->setToolTip("Select/modify mask");
    selectionButton->setCheckable(true);
    addWidget(selectionButton);

    m_activityButtonGroup->addButton(panButton, MaskFlags::PAN_ZOOM_MODE);
    m_activityButtonGroup->addButton(selectionButton, MaskFlags::SELECTION_MODE);

    //... Shape creation tools

    auto* roiButton = new QToolButton(this);
    roiButton->setIcon(QIcon(":/images/maskeditor_roi.svg"));
    roiButton->setToolTip("Create region of interest");
    roiButton->setCheckable(true);
    addWidget(roiButton);

    auto* rectangleButton = new QToolButton(this);
    rectangleButton->setIcon(QIcon(":/images/maskeditor_rectangle.svg"));
    rectangleButton->setToolTip("Create rectangle mask");
    rectangleButton->setCheckable(true);
    addWidget(rectangleButton);

    auto* polygonButton = new QToolButton(this);
    polygonButton->setIcon(QIcon(":/images/maskeditor_polygon.svg"));
    polygonButton->setToolTip("Create polygon mask");
    polygonButton->setCheckable(true);
    addWidget(polygonButton);

    auto* verticalLineButton = new QToolButton(this);
    verticalLineButton->setIcon(QIcon(":/images/maskeditor_verticalline.svg"));
    verticalLineButton->setToolTip("Create vertical line mask");
    verticalLineButton->setCheckable(true);
    addWidget(verticalLineButton);

    auto* horizontalLineButton = new QToolButton(this);
    horizontalLineButton->setIcon(QIcon(":/images/maskeditor_horizontalline.svg"));
    horizontalLineButton->setToolTip("Create horizontal line mask");
    horizontalLineButton->setCheckable(true);
    addWidget(horizontalLineButton);

    auto* ellipseButton = new QToolButton(this);
    ellipseButton->setIcon(QIcon(":/images/maskeditor_ellipse.svg"));
    ellipseButton->setToolTip("Create ellipse mask");
    ellipseButton->setCheckable(true);
    addWidget(ellipseButton);

    //    QToolButton *maskAllButton = new QToolButton(this);
    //    maskAllButton->setIcon(QIcon(":/images/maskeditor_maskall.svg"));
    //    maskAllButton->setToolTip("Create masked area covering whole detector plane\n"
    //                              "Will be placed beneath all masks. Only one instance is
    //                              allowed.");
    //    maskAllButton->setCheckable(true);
    //    addWidget(maskAllButton);

    m_activityButtonGroup->addButton(roiButton, MaskFlags::ROI_MODE);
    m_activityButtonGroup->addButton(rectangleButton, MaskFlags::RECTANGLE_MODE);
    m_activityButtonGroup->addButton(polygonButton, MaskFlags::POLYGON_MODE);
    m_activityButtonGroup->addButton(verticalLineButton, MaskFlags::VERTICAL_LINE_MODE);
    m_activityButtonGroup->addButton(horizontalLineButton, MaskFlags::HORIZONTAL_LINE_MODE);
    m_activityButtonGroup->addButton(ellipseButton, MaskFlags::ELLIPSE_MODE);
    //    m_activityButtonGroup->addButton(maskAllButton, MaskFlags::MASKALL_MODE);
    add_separator();

    //... Connect "Activity" button group

    connect(m_activityButtonGroup, &QButtonGroup::idClicked,
            [this](int) { emit activityChanged(currentActivity()); });

    //... Modifier tools

    ASSERT(actions);
    addAction(actions->bringToFrontAction());
    addAction(actions->sendToBackAction());
    add_separator();

    /*
    auto* maskValueGroup = new QButtonGroup(this);
    addWidget(new QLabel(" "));

    auto* maskTrueButton = new QRadioButton(this);
    maskTrueButton->setText("masked");
    maskTrueButton->setToolTip("Set mask to True (area is excluded from the simulation)");
    maskTrueButton->setCheckable(true);
    maskTrueButton->setChecked(true);
    addWidget(maskTrueButton);

    addWidget(new QLabel(" "));

    auto* maskFalseButton = new QRadioButton(this);
    maskFalseButton->setToolTip("Set mask to False (area stays in the simulation)");
    maskFalseButton->setCheckable(true);
    maskFalseButton->setText("active");
    addWidget(maskFalseButton);

    maskValueGroup->addButton(maskTrueButton, MaskFlags::MASK_ON);
    maskValueGroup->addButton(maskFalseButton, MaskFlags::MASK_OFF);

    connect(maskValueGroup, &QButtonGroup::idClicked,
            [this](int value) { emit maskValueChanged(MaskFlags::MaskValue(value)); });
    */

    //... Inspection tools

    auto* presentationButton = new QToolButton(this);
    presentationButton->setIcon(QIcon(":/images/maskeditor_lightbulb.svg"));
    presentationButton->setToolTip("Press and hold to see mask results.");
    addWidget(presentationButton);

    connect(presentationButton, &QToolButton::pressed, [this] { emit presentationChange(true); });
    connect(presentationButton, &QToolButton::released, [this] { emit presentationChange(false); });

    auto* propertyPanelButton = new QToolButton(this);
    propertyPanelButton->setIcon(QIcon(":/images/maskeditor_toolpanel.svg"));
    propertyPanelButton->setToolTip("Open panel with additional properties");
    addWidget(propertyPanelButton);

    connect(propertyPanelButton, &QToolButton::clicked, actions,
            &MaskEditorActions::propertyPanelRequest);
    add_separator();
}

void MaskToolbar::onChangeActivityRequest(MaskFlags::MaskMode mask_mode)
{
    setCurrentActivity(mask_mode);
    emit activityChanged(currentActivity());
}

void MaskToolbar::add_separator()
{
    addWidget(new QLabel(" "));
    addSeparator();
    addWidget(new QLabel(" "));
}

MaskFlags::MaskMode MaskToolbar::currentActivity() const
{
    return MaskFlags::EMaskMode(m_activityButtonGroup->checkedId());
}

void MaskToolbar::setCurrentActivity(MaskFlags::MaskMode mask_mode)
{
    m_activityButtonGroup->button(mask_mode)->setChecked(true);
}
