//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Setup/ProjectionToolbar.cpp
//! @brief     Implements class ProjectionToolbar.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Setup/ProjectionToolbar.h"
#include "GUI/View/Setup/ProjectionActions.h"
#include "GUI/View/Tool/mainwindow_constants.h"
#include <QLabel>
#include <QToolButton>

ProjectionToolbar::ProjectionToolbar(ProjectionActions* editorActions)
    : m_editorActions(editorActions)
    , m_activityButtonGroup(new QButtonGroup(this))
{
    setOrientation(Qt::Vertical);
    setStyleSheet("border-left:1px solid rgb(180,180,180);");
    setIconSize(QSize(GUI::Style::toolbar_icon_size, GUI::Style::toolbar_icon_size));

    //... Selection group

    auto* panButton = new QToolButton(this);
    panButton->setIcon(QIcon(":/images/hand-right.svg"));
    panButton->setToolTip("Pan/zoom mode (space)\n"
                          "Drag axes with the mouse, use mouse wheel to zoom in/out");
    panButton->setCheckable(true);
    panButton->setChecked(true);
    addWidget(panButton);

    auto* resetViewButton = new QToolButton(this);
    resetViewButton->setIcon(QIcon(":/images/camera-metering-center.svg"));
    resetViewButton->setToolTip("Reset view\nx,y,z axes range will be set to default");
    addWidget(resetViewButton);
    connect(resetViewButton, &QToolButton::clicked, m_editorActions,
            &ProjectionActions::resetViewRequest);

    add_separator();

    auto* selectionButton = new QToolButton(this);
    selectionButton->setIcon(QIcon(":/images/maskeditor_arrow.svg"));
    selectionButton->setToolTip(
        "Selection mode\nYou can select existing projections and move them around");
    selectionButton->setCheckable(true);
    addWidget(selectionButton);

    m_activityButtonGroup->addButton(panButton, MaskFlags::PAN_ZOOM_MODE);
    m_activityButtonGroup->addButton(selectionButton, MaskFlags::SELECTION_MODE);

    //... Shapes group

    auto* horizontalLineButton = new QToolButton(this);
    horizontalLineButton->setIcon(QIcon(":/images/maskeditor_horizontalline.svg"));
    horizontalLineButton->setToolTip(
        "Horizontal projections mode\nCreate projection along x-axis by clicking on color map");
    horizontalLineButton->setCheckable(true);
    addWidget(horizontalLineButton);

    auto* verticalLineButton = new QToolButton(this);
    verticalLineButton->setIcon(QIcon(":/images/maskeditor_verticalline.svg"));
    verticalLineButton->setToolTip(
        "Vertical projections mode\nCreate projection along y-axis by clicking on color map");
    verticalLineButton->setCheckable(true);
    addWidget(verticalLineButton);

    m_activityButtonGroup->addButton(verticalLineButton, MaskFlags::VERTICAL_LINE_MODE);
    m_activityButtonGroup->addButton(horizontalLineButton, MaskFlags::HORIZONTAL_LINE_MODE);

    connect(m_activityButtonGroup, &QButtonGroup::idClicked,
            [this] { emit activityChanged(currentActivity()); });

    add_separator();

    //... Extratools group

    auto* saveButton = new QToolButton(this);
    saveButton->setIcon(QIcon(":/images/maskeditor_save.svg"));
    saveButton->setToolTip("Save created projections in multi-column ASCII file.");
    addWidget(saveButton);
    connect(saveButton, &QToolButton::clicked, m_editorActions, &ProjectionActions::onSaveAction);
}

void ProjectionToolbar::onChangeActivityRequest(MaskFlags::MaskMode mask_mode)
{
    setCurrentActivity(mask_mode);
    emit activityChanged(currentActivity());
}

//! Change activity only if current activity is one of drawing mode (horizontal, vertical
//! projections drawing).
void ProjectionToolbar::onProjectionTabChange(MaskFlags::MaskMode mask_mode)
{
    if (currentActivity() == MaskFlags::HORIZONTAL_LINE_MODE
        || currentActivity() == MaskFlags::VERTICAL_LINE_MODE)
        onChangeActivityRequest(mask_mode);
}

void ProjectionToolbar::add_separator()
{
    addWidget(new QLabel(" "));
    addSeparator();
    addWidget(new QLabel(" "));
}

MaskFlags::MaskMode ProjectionToolbar::currentActivity() const
{
    return MaskFlags::EMaskMode(m_activityButtonGroup->checkedId());
}

void ProjectionToolbar::setCurrentActivity(MaskFlags::MaskMode mask_mode)
{
    int button_index = static_cast<int>(mask_mode);
    m_activityButtonGroup->button(button_index)->setChecked(true);
}
