#include "Device/Data/Datafield.h"

#include "Device/IO/IOFactory.h"
#include "GUI/Model/Beam/SourceItems.h"
#include "GUI/Model/Data/Data2DItem.h"
#include "GUI/Model/Detector/DetectorItem.h"
#include "GUI/Model/Device/DatafileItem.h"
#include "GUI/Model/Device/InstrumentItems.h"
#include "GUI/Model/Job/JobItem.h"
#include "GUI/Model/Mask/MaskItems.h"
#include "GUI/Model/Mask/MaskList.h"
#include "GUI/Model/Mask/MaskeditorListmodel.h"
#include "GUI/Model/Model/JobModel.h"
#include "Tests/GTestWrapper/google_test.h"
#include "Tests/Unit/GUI/Utils.h"
#include <QTest>

TEST(TestJobModel, nonXMLData)
{
    JobModel jobModel;

    // initial state
    EXPECT_EQ(jobModel.dataItems().size(), 0);

    // adding JobItem
    auto* jobItem = jobModel.createJobItem();
    auto* instrumentItem = new GISASInstrumentItem();
    jobItem->copyInstrumentIntoJob(instrumentItem);
    jobItem->createSimulatedDataItem();
    EXPECT_EQ(jobModel.dataItems().size(), 1);

    // adding empty DatafileItem to JobItem
    auto* realItem = jobItem->createDatafileItem();
    EXPECT_EQ(jobModel.dataItems().size(), 1);

    // create specular data in DatafileItem
    realItem->setTheDatafield(UTest::GUI::makeData1D(0.0));
    EXPECT_EQ(jobModel.dataItems().size(), 2);

    // checking data items of JobModel
    EXPECT_EQ(jobModel.dataItems().indexOf(jobItem->simulatedDataItem()), 0);
    EXPECT_EQ(jobModel.dataItems().indexOf(realItem->dataItem()), 1);

    // replacing the data inside DatafileItem with the data of the same dimensions
    realItem->setTheDatafield(UTest::GUI::makeData1D(2.0));
    EXPECT_EQ(jobModel.dataItems().size(), 2);

    // replacing the data inside DatafileItem with the data of different dimensions
    const Datafield data = UTest::GUI::makeData2D(3.0);
    EXPECT_THROW(dynamic_cast<DatafileItem*>(realItem)->setTheDatafield(data), std::runtime_error);

    EXPECT_EQ(jobModel.dataItems().size(), 2);
}

TEST(TestJobModel, saveNonXMLData)
{
    const QString projectDir("DatafieldIOService");
    UTest::GUI::create_dir(projectDir);

    const double value1(1.0), value2(2.0), value3(3.0);

    JobModel jobModel;

    // adding JobItem with instrument
    auto* jobItem = jobModel.createJobItem();
    auto* instrumentItem = new GISASInstrumentItem();
    jobItem->copyInstrumentIntoJob(instrumentItem);

    // create simulated data
    jobItem->createSimulatedDataItem();
    jobItem->simulatedDataItem()->setDatafield(UTest::GUI::makeData2D(value1));

    // create real data
    jobItem->createDatafileItem();
    jobItem->realItem()->setTheDatafield(UTest::GUI::makeData2D(value2));

    // update names
    jobItem->setJobName("name");
    jobItem->updateFileName();

    // save first time
    jobModel.writeDatafiles(projectDir);
    QTest::qSleep(10);

    // check existence of data on disk
    QString fname1 = "./" + projectDir + "/jobdata_name.int";
    QString fname2 = "./" + projectDir + "/refdata_name.int";
    EXPECT_TRUE(QFile::exists(fname1));
    EXPECT_TRUE(QFile::exists(fname2));

    // read data from disk, check it is the same
    std::unique_ptr<Datafield> dataOnDisk1(
        IO::readData2D(fname1.toStdString(), IO::Filetype2D::bornagain2D));
    std::unique_ptr<Datafield> dataOnDisk2(
        IO::readData2D(fname2.toStdString(), IO::Filetype2D::bornagain2D));
    EXPECT_TRUE(UTest::GUI::isTheSame(*dataOnDisk1, *jobItem->simulatedDataItem()->c_field()));
    EXPECT_TRUE(UTest::GUI::isTheSame(*dataOnDisk2, *jobItem->realItem()->dataItem()->c_field()));

    // modify data and save the project.
    jobItem->simulatedDataItem()->setDatafield(UTest::GUI::makeData2D(value3));
    jobModel.writeDatafiles(projectDir);
    QTest::qSleep(10);

    // ensure that the simulated data has been changed
    EXPECT_FALSE(UTest::GUI::isTheSame(*dataOnDisk1, *jobItem->simulatedDataItem()->c_field()));
    EXPECT_TRUE(UTest::GUI::isTheSame(*dataOnDisk2, *jobItem->realItem()->dataItem()->c_field()));

    // check that data on disk has changed
    dataOnDisk1.reset(IO::readData2D(fname1.toStdString(), IO::Filetype2D::bornagain2D));
    EXPECT_TRUE(UTest::GUI::isTheSame(*dataOnDisk1, *jobItem->simulatedDataItem()->c_field()));

    // rename job and check that file on disk changed the name
    jobItem->setJobName("new_job");
    jobItem->updateFileName();
    jobModel.writeDatafiles(projectDir);
    QTest::qSleep(10);

    // check existence of new files on disk
    QString fname1new = "./" + projectDir + "/jobdata_new_job.int";
    QString fname2new = "./" + projectDir + "/refdata_new_job.int";
    EXPECT_TRUE(QFile::exists(fname1new));
    EXPECT_TRUE(QFile::exists(fname2new));

    // check that file with old name was removed
    EXPECT_FALSE(QFile::exists(fname1));
    EXPECT_FALSE(QFile::exists(fname2));
}

TEST(TestJobModel, masksToDetector)
{
    int nx = 7;
    int ny = 5;

    // create JobItem
    JobItem jobItem;
    auto* instrumentItem = new GISASInstrumentItem;
    instrumentItem->detectorItem()->phiAxis().setNbins(nx);
    instrumentItem->detectorItem()->alphaAxis().setNbins(ny);
    jobItem.copyInstrumentIntoJob(instrumentItem);

    // create simulated data (may be empty)
    jobItem.createSimulatedDataItem();

    // create external 2d data
    DatafileItem realItem;
    realItem.setTheDatafield(UTest::GUI::makeData2D(1.0, nx, -4.5, 2.5, ny, 0., 5.0));

    // create mask (M - masked pixel)
    //     _ _ _ _ _ _ _
    // 4  |_|_|_|_|_|_|_|
    // 3  |_|_|_|_|M|M|M|
    // 2  |_|_|_|_|M|M|M|
    // 1  |_|_|_|_|M|M|M|
    // 0  |_|_|_|_|_|_|_|
    //
    //     0 1 2 3 4 5 6

    RectangleItem mask;
    mask.setXLow(-0.5);
    mask.setXUp(2.5);
    mask.setYLow(1.0);
    mask.setYUp(4.0);

    // add mask to DatafileItem
    MaskList* container = realItem.data2DItem()->getOrCreateMaskModel()->maskContItem();
    container->addMaskItem(&mask);

    // add DatafileItem to job
    jobItem.copyDatafileItemIntoJob(&realItem);

    // check DatafileItem data before applying mask
    const auto* field_before = jobItem.realItem()->data2DItem()->c_field();
    for (size_t i = 0; i < field_before->size(); i++)
        EXPECT_EQ(field_before->valAt(i), 1.0);

    // adjust detector and DatafileItem to each other, copy and apply masks
    jobItem.adjustReaDataToJobInstrument();

    // check DatafileItem data after applying mask
    const auto* field_after = jobItem.realItem()->data2DItem()->c_field();
    auto flat_i = [nx](int ix, int iy) { return nx * iy + ix; };

    for (int ix = 0; ix < nx; ix++)
        for (int iy = 0; iy < ny; iy++)
            if ((4 <= ix && ix <= 6) && (1 <= iy && iy <= 4))
                EXPECT_EQ(field_after->valAt(flat_i(ix, iy)), 0.0); // masked
            else
                EXPECT_EQ(field_after->valAt(flat_i(ix, iy)), 1.0);
}
